{ File   : Shared.pas
  Author : Deniz Oezmen
  Created: 2007-07-22
  Changed: 2007-07-26

  IMA ADPCM format plugin manager.
}
unit Shared;

interface

type
  // IMA ADPCM algorithm interface types
  TIMA_ADPCM_State = record
    PrevSample: Smallint; // last predicted sample
    Index     : Byte;     // index into step table
  end;

  TIMA_ADPCM_States = array of TIMA_ADPCM_State;

  TIMA_ADPCM_SampleBuffer = array of Smallint;

  TIMA_ADPCM_CodeBuffer = array of Byte;

  // plugin interface function prototypes
  TCorrectParams = procedure (var BlockSize, SampPerBlock,
    NumChannels: Longint);
  TGetCompressedSize = function (BlockSize, SampPerBlock, NumChannels,
    NumSamples: Longint): Longint;
  TGetDecompressedSize = function (BlockSize, SampPerBlock, NumChannels,
    CodeSize: Longint): Longint;
  TCompressToBuffer = function (const SampBuffer: TIMA_ADPCM_SampleBuffer;
    var SampBufIndex: Longint; const CodeBuffer: TIMA_ADPCM_CodeBuffer;
    var CodeBufIndex: Longint; const States: TIMA_ADPCM_States;
    BlockSize, SampPerBlock, NumChannels: Longint): Longint;
  TDecompressFromBuffer = function (const SampBuffer: TIMA_ADPCM_SampleBuffer;
    var SampBufIndex: Longint; const CodeBuffer: TIMA_ADPCM_CodeBuffer;
    var CodeBufIndex: Longint; const States: TIMA_ADPCM_States;
    BlockSize, SampPerBlock, NumChannels: Longint): Longint;

  TPluginFunctions = record
    CorrectParams       : TCorrectParams;
    GetCompressedSize   : TGetCompressedSize;
    GetDecompressedSize : TGetDecompressedSize;
    CompressToBuffer    : TCompressToBuffer;
    DecompressFromBuffer: TDecompressFromBuffer;
  end;

// published plugin management functions
function RegisterPlugin(const PluginGUID: TGUID; const PluginDesc: string;
  const PluginFunctions: TPluginFunctions): Longint;
procedure UnregisterPlugin(const PluginGUID: TGUID);
function GetPluginFunctions(const PluginGUID: TGUID;
  out PluginFunctions: TPluginFunctions): Boolean;

// published IMA ADPCM algorithm functions
function EncodeSample(var State: TIMA_ADPCM_State; Sample: Smallint): Byte;
function DecodeSample(var State: TIMA_ADPCM_State; Code: Byte): Smallint;
function GetBestIndex(Sample, PrevSample: Smallint): Byte;

implementation

uses
  SysUtils;

const
  StepTab: array[0..88] of Smallint =
    ($0007, $0008, $0009, $000a, $000b, $000c, $000d, $000e,
     $0010, $0011, $0013, $0015, $0017, $0019, $001c, $001f,
     $0022, $0025, $0029, $002d, $0032, $0037, $003c, $0042,
     $0049, $0050, $0058, $0061, $006b, $0076, $0082, $008f,
     $009d, $00ad, $00be, $00d1, $00e6, $00fd, $0117, $0133,
     $0151, $0173, $0198, $01c1, $01ee, $0220, $0256, $0292,
     $02d4, $031c, $036c, $03c3, $0424, $048e, $0502, $0583,
     $0610, $06ab, $0756, $0812, $08e0, $09c3, $0abd, $0bd0,
     $0cff, $0e4c, $0fba, $114c, $1307, $14ee, $1706, $1954,
     $1bdc, $1ea5, $21b6, $2515, $28ca, $2cdf, $315b, $364b,
     $3bb9, $41b2, $4844, $4f7e, $5771, $602f, $69ce, $7462,
     $7fff);
  IndexTab: array[0..15] of Smallint =
    ( -1, -1, -1, -1,  2,  4,  6,  8,
      -1, -1, -1, -1,  2,  4,  6,  8);

type
  TPlugin = record
    PluginGUID     : TGUID;
    PluginDesc     : string;
    PluginFunctions: TPluginFunctions;
    Registered     : Boolean;
  end;

var
  Plugins: array of TPlugin;

function RegisterPlugin(const PluginGUID: TGUID; const PluginDesc: string;
  const PluginFunctions: TPluginFunctions): Longint;
var
  i: Longint;
begin
  Result := -1;

  for i := 0 to High(Plugins) do
    if CompareMem(@Plugins[i].PluginGUID, @PluginGUID, SizeOf(TGUID)) and
      Plugins[i].Registered then
    begin
      Result := i;
      Break
    end;

  if Result <> -1 then
    Exit;

  for i := 0 to High(Plugins) do
    if not Plugins[i].Registered then
      Result := i;

  if Result = - 1 then
  begin
    SetLength(Plugins, Length(Plugins) + 1);
    Result := High(Plugins)
  end;

  Plugins[Result].PluginGUID := PluginGUID;
  Plugins[Result].PluginDesc := PluginDesc;
  Plugins[Result].PluginFunctions := PluginFunctions;
  Plugins[Result].Registered := True
end;

procedure UnregisterPlugin(const PluginGUID: TGUID);
var
  i,
  LastAssigned: Longint;
begin
  for i := 0 to High(Plugins) do
    if CompareMem(@Plugins[i].PluginGUID, @PluginGUID, SizeOf(TGUID)) then
    begin
      FillChar(Plugins[i].PluginGUID, SizeOf(TGUID), 0);
      Plugins[i].PluginFunctions.CorrectParams := nil;
      Plugins[i].PluginFunctions.GetCompressedSize := nil;
      Plugins[i].PluginFunctions.GetDecompressedSize := nil;
      Plugins[i].PluginFunctions.CompressToBuffer := nil;
      Plugins[i].PluginFunctions.DecompressFromBuffer := nil;
      Plugins[i].Registered := False;

      Break
    end;

  LastAssigned := -1;
  for i := 0 to High(Plugins) do
    if Plugins[i].Registered then
      LastAssigned := i;

  SetLength(Plugins, LastAssigned + 1)
end;

function GetPluginFunctions(const PluginGUID: TGUID;
  out PluginFunctions: TPluginFunctions): Boolean;
var
  i: Longint;
begin
  Result := False;

  for i := 0 to High(Plugins) do
    if CompareMem(@Plugins[i].PluginGUID, @PluginGUID, SizeOf(TGUID)) and
      Plugins[i].Registered then
    begin
      Result := True;
      PluginFunctions := Plugins[i].PluginFunctions
    end
end;

function EncodeSample(var State: TIMA_ADPCM_State; Sample: Smallint): Byte;
var
  Delta,
  StepSize,
  NewSample,
  NewIndex : Longint;
begin
  Result := 0;

  // get delta for predictor
  Delta := Longint(Sample) - Longint(State.PrevSample);
  NewSample := Delta;
  StepSize := StepTab[State.Index];

  // quantizer step via delta
  if Delta < 0 then
  begin
    Delta := -Delta;
    Result := $08
  end;
  if Delta >= StepSize then
  begin
    Dec(Delta, StepSize);
    Result := Result or $04;
  end;
  StepSize := StepSize shr 1;
  if Delta >= StepSize then
  begin
    Dec(Delta, StepSize);
    Result := Result or $02;
  end;
  StepSize := StepSize shr 1;
  if Delta >= StepSize then           
  begin
    Dec(Delta, StepSize);
    Result := Result or $01
  end;
  StepSize := StepSize shr 1;
  Dec(Delta, StepSize);

  // construct predicted sample
  if NewSample < 0 then
    Delta := -Delta;
  NewSample := Longint(State.PrevSample) + NewSample - Delta;
  if NewSample > 32767 then
    NewSample := 32767
  else if NewSample < -32768 then
    NewSample := -32768;

  // calculate new step table index
  NewIndex := State.Index + IndexTab[Result];
  if NewIndex > 88 then
    NewIndex := 88
  else if NewIndex < 0 then
    NewIndex := 0;

  // update compressor state
  State.PrevSample := NewSample;
  State.Index := NewIndex
end;

function DecodeSample(var State: TIMA_ADPCM_State; Code: Byte): Smallint;
var
  Sample,
  Index,
  Delta,
  StepSize : Longint;
begin
  // expand parameters' value range for computations
  Sample := Longint(State.PrevSample);
  Index := Longint(State.Index);

  // calculate difference to predicted sample
  Delta := 0;
  StepSize := StepTab[Index];
  if Code and $04 <> 0 then
    Inc(Delta, StepSize);
  if Code and $02 <> 0 then
    Inc(Delta, StepSize shr 1);
  if Code and $01 <> 0 then
    Inc(Delta, StepSize shr 2);
  Inc(Delta, StepSize shr 3);
  if Code and $08 <> 0 then
    Delta := -Delta;

  // calculate new sample
  Inc(Sample, Delta);
  if Sample > 32767 then
    Sample := 32767
  else if Sample < -32768 then
    Sample := -32768;

  // calculate new step table index
  Inc(Index, IndexTab[Code]);
  if Index > 88 then
    Index := 88
  else if Index < 0 then
    Index := 0;

  // write values back to parameters
  State.PrevSample := Smallint(Sample);
  State.Index := Byte(Index);

  // also return decoded sample for convenience
  Result := State.PrevSample
end;

function GetBestIndex(Sample, PrevSample: Smallint): Byte;
var
  i,
  Delta,
  MinRemain,
  StepSize : Longint;
begin
  MinRemain := MaxInt;
  Result := 0;

  for i := 0 to 88 do
  begin
    Delta := Longint(Sample) - Longint(PrevSample);
    StepSize := StepTab[i];

    if Delta < 0 then
      Delta := -Delta;
    if Delta >= StepSize then
      Dec(Delta, StepSize);
    StepSize := StepSize shr 1;
    if Delta >= StepSize then
      Dec(Delta, StepSize);
    StepSize := StepSize shr 1;
    if Delta >= StepSize then
      Dec(Delta, StepSize);
    StepSize := StepSize shr 1;
      Dec(Delta, StepSize);

    if Abs(Delta) < MinRemain then
    begin
      MinRemain := Abs(Delta);
      Result := i
    end
  end
end;

initialization
  SetLength(Plugins, 0);

finalization
  Finalize(Plugins);
  SetLength(Plugins, 0);
  Plugins := nil;

end.
